from scipy import interpolate

from .basetest import BaseTest
from ..calculators import AsymptoticCalculator
from ..parameters import POI
from ..exceptions import POIRangeError


class UpperLimit(BaseTest):
    def __init__(self, calculator, poinull, poialt, qtilde=False):
        """Class for upper limit calculation.

            Args:
                calculator (`sktats.hypotests.BaseCalculator`): calculator to use for computing the pvalues
                poinull (List[`hypotests.POI`]): parameters of interest for the null hypothesis
                poialt (List[`hypotests.POI`]): parameters of interest for the alternative hypothesis
                qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                    the $$q$$ test statistic

            Example with `zfit`:
                >>> import numpy as np
                >>> import zfit
                >>> from zfit.loss import ExtendedUnbinnedNLL
                >>> from zfit.minimize import MinuitMinimizer

                >>> bounds = (0.1, 3.0)
                >>> zfit.Space('x', limits=bounds)

                >>> bkg = np.random.exponential(0.5, 300)
                >>> peak = np.random.normal(1.2, 0.1, 10)
                >>> data = np.concatenate((bkg, peak))
                >>> data = data[(data > bounds[0]) & (data < bounds[1])]
                >>> N = data.size
                >>> data = zfit.data.Data.from_numpy(obs=obs, array=data)

                >>> lambda_ = zfit.Parameter("lambda", -2.0, -4.0, -1.0)
                >>> Nsig = zfit.Parameter("Ns", 20., -20., N)
                >>> Nbkg = zfit.Parameter("Nbkg", N, 0., N*1.1)
                >>> signal = Nsig * zfit.pdf.Gauss(obs=obs, mu=1.2, sigma=0.1)
                >>> background = Nbkg * zfit.pdf.Exponential(obs=obs, lambda_=lambda_)
                >>> loss = ExtendedUnbinnedNLL(model=[signal + background], data=[data])

                >>> from hepstats.hypotests.calculators import AsymptoticCalculator
                >>> from hepstats.hypotests import UpperLimit
                >>> from hepstats.hypotests.parameters import POI, POIarray

                >>> calculator = AsymptoticCalculator(loss, MinuitMinimizer())
                >>> poinull = POIarray(Nsig, np.linspace(0.0, 25, 20))
                >>> poialt = POI(Nsig, 0)
                >>> ul = UpperLimit(calculator, [poinull], [poialt])
                >>> ul.upperlimit(alpha=0.05, CLs=True)
                Observed upper limit: Nsig = 15.725784747406346
                Expected upper limit: Nsig = 11.927442041887158
                Expected upper limit +1 sigma: Nsig = 16.596396280677116
                Expected upper limit -1 sigma: Nsig = 8.592750403611896
                Expected upper limit +2 sigma: Nsig = 22.24864429383046
                Expected upper limit -2 sigma: Nsig = 6.400549971360598
        """

        super(UpperLimit, self).__init__(calculator, poinull, poialt)

        self._qtilde = qtilde

    @property
    def qtilde(self):
        """
        Returns True if qtilde test statistic is used, else False.
        """
        return self._qtilde

    def pvalues(self, CLs=True):
        """
        Returns p-values scanned for the values of the parameters of interest
        in the null hypothesis.

        Args:
            CLs (bool, optional): if `True` uses pvalues as $$p_{cls}=p_{null}/p_{alt}=p_{clsb}/p_{clb}$$
                else as $$p_{clsb} = p_{null}$

        Returns:
            pvalues (Dict): CLsb, CLs, expected (+/- sigma bands) p-values
        """
        pvalue_func = self.calculator.pvalue

        pnull, palt = pvalue_func(poinull=self.poinull, poialt=self.poialt, qtilde=self.qtilde, onesided=True)

        pvalues = {"clsb": pnull, "clb": palt}

        sigmas = [0.0, 1.0, 2.0, -1.0, -2.0]

        exppvalue_func = self.calculator.expected_pvalue

        result = exppvalue_func(poinull=self.poinull, poialt=self.poialt, nsigma=sigmas, CLs=CLs,
                                qtilde=self.qtilde, onesided=True)

        pvalues["expected"] = result[0]
        pvalues["expected_p1"] = result[1]
        pvalues["expected_p2"] = result[2]
        pvalues["expected_m1"] = result[3]
        pvalues["expected_m2"] = result[4]

        pvalues["cls"] = pnull / palt

        return pvalues

    def upperlimit(self, alpha=0.05, CLs=True, printlevel=1):
        """
        Returns the upper limit of the parameter of interest.

        Args:
            alpha (float, default=0.05): significance level
            CLs (bool, optional): if `True` uses pvalues as $$p_{cls}=p_{null}/p_{alt}=p_{clsb}/p_{clb}$$
                else as $$p_{clsb} = p_{null}$
            printlevel (int, default=1): if > 0 print the result

        Returns:
            limits (Dict): observed, expected (+/- sigma bands) upper limits

        """

        poinull = self.poinull

        # create a filter for -1 and -2 sigma expected limits
        bestfit = self.calculator.bestfit.params[poinull.parameter]["value"]
        filter = poinull.values >= bestfit

        if CLs:
            observed_key = "cls"
        else:
            observed_key = "clsb"

        if isinstance(self.calculator, AsymptoticCalculator):
            to_interpolate = [observed_key]
        else:
            to_interpolate = [observed_key] + [f"expected{i}" for i in ["", "_p1", "_m1", "_p2", "_m2"]]

        limits = {}
        for k in to_interpolate:

            pvalues = self.pvalues(CLs)[k]
            values = poinull.values

            if k == observed_key:
                k = "observed"
                pvalues = pvalues[filter]
                values = values[filter]

            if min(pvalues) > alpha:
                if k in ["expected", "observed"]:
                    msg = f"The minimum of the scanned p-values is {min(pvalues)} which is larger than the"
                    msg += f" confidence level alpha = {alpha}. Try to increase the maximum POI value."
                    raise POIRangeError(msg)
                else:
                    limits[k] = None
                    continue

            tck = interpolate.splrep(values, pvalues-alpha, s=0)
            root = interpolate.sproot(tck)

            if len(root) > 1:
                root = root[0]

            try:
                limits[k] = float(root)
            except TypeError:
                limits[k] = None

        if isinstance(self.calculator, AsymptoticCalculator):
            poiul = POI(poinull.parameter, limits["observed"])
            exppoi_func = self.calculator.expected_poi
            sigmas = [0.0, 1.0, -1.0, 2.0, -2.0]

            results = exppoi_func(poinull=poiul, poialt=self.poialt, nsigma=sigmas, alpha=alpha, CLs=CLs)
            keys = [f"expected{i}" for i in ["", "_p1", "_m1", "_p2", "_m2"]]

            for r, k in zip(results, keys):
                limits[k] = float(r)

        if printlevel > 0:
            print(f"\nObserved upper limit: {poinull.name} = {limits['observed']}")
            print(f"Expected upper limit: {poinull.name} = {limits['expected']}")
            for sigma in ["+1", "-1", "+2", "-2"]:
                key = sigma.replace("+", "p").replace("-", "m")
                print(f"Expected upper limit {sigma} sigma: {poinull.name} = {limits[f'expected_{key}']}")

        return limits



"""
Specific exceptions for the `hypotests` submodule
"""


class POIRangeError(Exception):
    """Exception class non adequate POI scan range"""

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ParameterNotFound(Exception):
    """Exception class raised if a parameter with a given name is not found"""

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class FormatError(Exception):
    """Exception class raised when unexpected yaml format are read"""

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)



import asdf
import os
import numpy as np
import warnings

from .parameters import POI, POIarray
from .exceptions import ParameterNotFound, FormatError
from .fitutils.utils import pll
from .fitutils.sampling import base_sampler, base_sample
from .hypotests_object import ToysObject

"""
Module defining the classes to perform and store the results of toy experiments.

Acronyms used in the code:
    * nll = negative log-likehood, the likehood being the `loss` attribute of a calculator;
"""


class ToyResult(object):

    def __init__(self, poigen, poieval):
        """
        Class to store the results of toys generated for a given value of a POI.
        The best fit value of the POI, the NLL evaluate at the best fit, and the NLL evaluated
        at several values of the POI are stored. The results can serialized using the `to_dict` method.

        Args:
            poigen (POI): POI used to generate the toys
            poieval (POIarray): POI values to evaluate the loss function
        """

        if not isinstance(poigen, POI):
            raise TypeError("A `hypotests.parameters.POI` is required for poigen.")
        if not isinstance(poieval, POIarray):
            raise TypeError("A `hypotests.parameters.POIarray` is required for poieval.")

        self._poigen = poigen
        self._bestfit = np.array([])
        self._nll_bestfit = np.array([])
        self._poieval = poieval
        self._nlls = {p: np.array([]) for p in poieval}

    @property
    def poigen(self):
        """
        Returns the POI used to generate the toys.
        """
        return self._poigen

    @property
    def bestfit(self):
        """
        Returns the best fitted values of the POI for each toys.
        """
        return self._bestfit

    @property
    def poieval(self):
        """
        Returns the scanned POIarray.
        """
        return self._poieval

    @property
    def nll_bestfit(self):
        """
        Returns the NLL evaluated at the best fitted values of the POI for each toys.
        """
        return self._nll_bestfit

    @property
    def nlls(self):
        """
        Returns the NLL evaluated at the poigeval values of the POI for each toys.
        """
        return self._nlls

    @property
    def ntoys(self):
        """
        Returns the number of toys.
        """
        return len(self.bestfit)

    def add_entries(self, bestfit, nll_bestfit, nlls):
        """
        Add new result entries.

        Args:
            * bestfit (`numpy.array`): best fitted values of the POI
            * nll_bestfit (`numpy.array`): NLL  evaluated at the best fitted values of the POI
            * nlls (Dict(`POI`, `numpy.array`)): NLL  evaluated at the best fitted values of the POI
        """
        if not all(k in nlls.keys() for k in self.poieval):
            missing_keys = [k for k in self.poieval if k not in nlls.keys()]
            raise ValueError(f"NLLs values for {missing_keys} are missing.")

        nentries = bestfit.size
        assert nll_bestfit.size == nentries
        assert all(a.size == nentries for a in nlls.values())

        self._bestfit = np.concatenate([self.bestfit, bestfit])
        self._nll_bestfit = np.concatenate([self.nll_bestfit, nll_bestfit])

        self._nlls = {p: np.concatenate([v, nlls[p]]) for p, v in self.nlls.items()}

    def to_dict(self):
        """
        Returns dictionary of the toy results.

        Keys:
            poi: name of the parameter of interest
            genvalues: fixed vale of the poi used to generate the toys
            evalvalues: values to evaluate the NLL
            bestfit: array of best fitted values of the poi for each toy
            nlls: dictionary of NLL values for each value in `evalvalues` and best fit
        """
        ret = {"poi": self.poigen.name, "bestfit": self.bestfit}
        ret["nlls"] = {n.value: nll for n, nll in self.nlls.items()}
        ret["genvalue"] = self.poigen.value
        ret["evalvalues"] = self.poieval.values
        ret["nlls"]["bestfit"] = self.nll_bestfit
        return ret


class FitFailuresWarning(UserWarning):
    pass


class ToysManager(ToysObject):

    def __init__(self, input, minimizer, sampler=base_sampler, sample=base_sample):
        """Class handling the toy generation and fit, results are stored in `ToyResult` instances stored
            themselves in a dictionary.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function
                ntoysnull (int, default=100): minimum number of toys to generate for the null hypothesis
                ntoysalt (int, default=100): minimum number of toys to generate for the alternative hypothesis
                sampler : function used to create sampler with models, number of events and
                    floating parameters in the sample. Default is `hepstats.fitutils.sampling.base_sampler`.
                sample : function used to get samples from the sampler.
                    Default is `hepstats.fitutils.sampling.base_sample`.
        """

        super(ToysManager, self).__init__(input=input, minimizer=minimizer, sampler=sampler, sample=sample)
        self._toys = {}

    def get_toyresult(self, poigen, poieval):
        """
        Getter function.

        Args:
            poigen (POI): POI used to generate the toys
            poieval (POIarray, optional): POI values to evaluate the loss function

        Returns:
            `ToyResult`
        """

        index = (poigen, poieval)

        if index not in self.keys():
            for k in self.keys():
                poigen_k, poieval_k = k
                if poigen_k != poigen:
                    continue
                if np.isin(poieval.values, k[-1].values).all():
                    index = k
                    break

        return self._toys[index]

    def add_toyresult(self, toy):
        """
        Add ToyResult to the manager.

        Args:
            toy: (`ToyResult`)
        """
        if not isinstance(toy, ToyResult):
            raise TypeError("A `hypotests.toyutils.ToyResult` is required for toy.")

        index = (toy.poigen, toy.poieval)

        self._toys[index] = toy

    def ntoys(self, poigen, poieval):
        """
        Return the number of toys generated from given value of a POI, and scanned/evaluated for given values_equal
        of the same POI.

        Args:
            poigen (POI): POI used to generate the toys
            poieval (POIarray, optional): POI values to evaluate the loss function

        Returns:
            int
        """
        try:
            return self.get_toyresult(poigen, poieval).ntoys
        except KeyError:
            return 0

    def generate_and_fit_toys(self, ntoys, poigen, poieval, printfreq=0.2):
        """
        Generate and fit toys for at a given POI (poigen). The toys are then fitted, and the likelihood
        is profiled at the values of poigen and poieval.

        Args:
            ntoys (int): number of toys to generate
            poigen (POI): POI used to generate the toys
            poieval (POIarray, optional): POI values to evaluate the loss function
            printfreq: print frequency of the toys generation
        """

        self.set_dependents_to_bestfit()

        minimizer = self.minimizer
        param = poigen.parameter

        toys_loss = self.toys_loss(poigen.name)
        sampler = toys_loss.data

        bestfit = np.empty(ntoys)
        nll_bestfit = np.empty(ntoys)
        nlls = {p: np.empty(ntoys) for p in poieval}

        printfreq = ntoys * printfreq

        samples = self.sample(sampler, int(ntoys*1.2), poigen)

        try:
            toysresult = self.get_toyresult(poigen, poieval)
            poieval = toysresult.poieval
        except KeyError:
            toysresult = ToyResult(poigen, poieval)
            self.add_toyresult(toysresult)

        nfailures = 0
        ntrials = 0

        for i in range(ntoys):
            ntrials += 1.
            converged = False
            toprint = i % printfreq == 0
            while converged is False:
                try:
                    next(samples)
                except StopIteration:
                    to_gen = ntoys - i
                    samples = self.sample(sampler, int(to_gen*1.2), poigen)
                    next(samples)
                     
                for minimize_trial in range(2):
                    try:
                        minimum = minimizer.minimize(loss=toys_loss)
                        converged = minimum.converged
                        if converged:
                            break
                    except RuntimeError:
                        converged = False
                        break

                if not converged:
                    self.set_dependents_to_bestfit()
                    nfailures += 1
                    if nfailures > 0.15 * ntrials and ntrials > 10:
                        msg = f"{nfailures} out of {ntrials} fits failed or didn't converge."
                        warnings.warn(msg, FitFailuresWarning)
                    continue

                bestfit[i] = minimum.params[param]["value"]
                nll_bestfit[i] = pll(minimizer, toys_loss, POI(param, bestfit[i]))

                for p in poieval:
                    nlls[p][i] = pll(minimizer, toys_loss, p)

            if toprint:
                print("{0} toys generated, fitted and scanned!".format(i))

            if i > ntoys:
                break
            i += 1

        if (poigen, poieval) not in self.keys():
            toysresult = ToyResult(p, poieval)
            self.add_toyresult(toysresult)
        else:
            toysresult = self.get_toyresult(poigen, poieval)

        toysresult.add_entries(bestfit=bestfit, nll_bestfit=nll_bestfit, nlls=nlls)

    def keys(self):
        return self._toys.keys()

    def values(self):
        return self._toys.values()

    def toyresults_to_dict(self):
        """
        Returns a list of all the toy results converted into dictionnaries
        """
        return [v.to_dict() for v in self.values()]

    def to_yaml(self, filename):
        """
        Save the toys into a yaml file under the key `toys`.

        Args:
            filename (str)
        """
        if os.path.isfile(filename):
            tree = asdf.open(filename).tree
        else:
            tree = {}

        tree["toys"] = self.toyresults_to_dict()
        af = asdf.AsdfFile(tree)
        af.write_to(filename)

    def toysresults_from_yaml(self, filename):
        """
        Extract toy results from a yaml file.

        Args:
            filename (str)

        Returns:
            list(`ToyResult`)
        """
        ret = []
        try:
            toys = asdf.open(filename).tree["toys"]
        except KeyError:
            raise FormatError(f"The key `toys` is not found in {filename}.")

        for t in toys:
            poiparam = None
            for p in self.loss.get_dependents():
                if t["poi"] == p.name:
                    poiparam = p

            if poiparam is None:
                raise ParameterNotFound(f"Parameter with name {t['poi']} is not found.")

            poigen = POI(poiparam, t["genvalue"])
            poieval = POIarray(poiparam, np.asarray(t["evalvalues"]))

            bestfit = t["bestfit"]
            nll_bestfit = t["nlls"]["bestfit"]
            nlls = {p: t["nlls"][p.value] for p in poieval}

            t = ToyResult(poigen, poieval)
            t.add_entries(bestfit=bestfit, nll_bestfit=nll_bestfit, nlls=nlls)
            ret.append(t)

        return ret

    @classmethod
    def from_yaml(cls, filename, input, minimizer, sampler=base_sampler, sample=base_sample):
        """
        Read the toys from a yaml file.

        Args:
            filename (str)
            input : loss or fit result
            minimizer : minimizer to use to find the minimum of the loss function
            ntoysnull (int, default=100): minimum number of toys to generate for the null hypothesis
            ntoysalt (int, default=100): minimum number of toys to generate for the alternative hypothesis
            sampler : function used to create sampler with models, number of events and
                floating parameters in the sample. Default is `hepstats.fitutils.sampling.base_sampler`.
            sample : function used to get samples from the sampler.
                Default is `hepstats.fitutils.sampling.base_sample`.

        Returns
            `ToysManager`
        """

        toyscollection = cls(input, minimizer, sampler, sample)
        toysresults = toyscollection.toysresults_from_yaml(filename)

        for t in toysresults:
            toyscollection.add_toyresult(t)

        return toyscollection



import numpy as np
from scipy.stats import norm

from .basecalculator import BaseCalculator
from ..fitutils.utils import eval_pdf, array2dataset, pll
from ..parameters import POI, POIarray


def generate_asimov_hist(model, params, nbins=100):
    """ Generate the Asimov histogram using a model and dictionary of parameters.

        Args:
            model : model used to generate the dataset
            params (Dict) : values of the parameters of the models
            nbins (int, optional) : number of bins

        Returns:
             (`np.array`, `np.array`) : hist, bin_edges

        Example with `zfit`:
            >>> obs = zfit.Space('x', limits=(0.1, 2.0))
            >>> mean = zfit.Parameter("mu", 1.2)
            >>> sigma = zfit.Parameter("sigma", 0.1)
            >>> model = zfit.pdf.Gauss(obs=obs, mu=mean, sigma=sigma)
            >>> hist, bin_edges = generate_asimov_hist(model, {"mean": 1.2, "sigma": 0.1})
    """

    space = model.space
    bounds = space.limit1d
    bin_edges = np.linspace(*bounds, nbins+1)
    bin_centers = bin_edges[0: -1] + np.diff(bin_edges)/2

    hist = eval_pdf(model, bin_centers, params)
    hist *= (space.area() / nbins)

    return hist, bin_edges


class AsymptoticCalculator(BaseCalculator):
    """
    Class for asymptotic calculators. Can be used only with one parameter
    of interest.

    See G. Cowan, K. Cranmer, E. Gross and O. Vitells: Asymptotic formulae for
    likelihood- based tests of new physics. Eur. Phys. J., C71:1–19, 2011
    """

    def __init__(self, input, minimizer, asimov_bins=100):
        """Asymptotic calculator class.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function
                asimov_bins (Optional, int) : number of bins of the Asimov dataset

            Example with `zfit`:
                >>> import zfit
                >>> from zfit.core.loss import UnbinnedNLL
                >>> from zfit.minimize import MinuitMinimizer

                >>> obs = zfit.Space('x', limits=(0.1, 2.0))
                >>> data = zfit.data.Data.from_numpy(obs=obs, array=np.random.normal(1.2, 0.1, 10000))
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> sigma = zfit.Parameter("sigma", 0.1)
                >>> model = zfit.pdf.Gauss(obs=obs, mu=mean, sigma=sigma)
                >>> loss = UnbinnedNLL(model=[model], data=[data], fit_range=[obs])

                >>> calc = AsymptoticCalculator(input=loss, minimizer=MinuitMinimizer(), nbins=100)
        """

        super(AsymptoticCalculator, self).__init__(input, minimizer)
        self._asimov_bins = asimov_bins
        self._asimov_dataset = {}
        self._asimov_loss = {}
        # cache of nll values computed with the asimov dataset
        self._asimov_nll = {}

    @staticmethod
    def check_pois(pois):
        """
        Check if the parameters of interest are all `hepstats.parameters.POI/POIarray` instances.
        """

        msg = "POI/POIarray is required."
        if not isinstance(pois, POIarray):
            raise TypeError(msg)
        if pois.ndim > 1:
            msg = "Tests using the asymptotic calculator can only be used with one parameter of interest."
            raise NotImplementedError(msg)

    def asimov_dataset(self, poi) -> (np.array, np.array):
        """ Generate the Asimov dataset for a given alternative hypothesis.

            Args:
                poi (`POI`): parameter of interest of the alternative hypothesis

            Returns:
                 Dataset

            Example with `zfit`:
                >>> poialt = POI(mean, 1.2)
                >>> dataset = calc.asimov_dataset([poialt])

        """

        if poi not in self._asimov_dataset.keys():
            model = self.model
            data = self.data
            minimizer = self.minimizer
            oldverbose = minimizer.verbosity
            minimizer.verbosity = 5

            poiparam = poi.parameter
            poivalue = poi.value

            msg = "\nGet fit best values for nuisance parameters for the"
            msg += " alternative hypothesis!"
            print(msg)

            with poiparam.set_value(poivalue):
                poiparam.floating = False
                minimum = minimizer.minimize(loss=self.loss)
                poiparam.floating = True

            minimizer.verbosity = oldverbose

            values = minimum.params
            values[poiparam] = {"value": poivalue}

            asimov_data = []
            for i, ad in enumerate([generate_asimov_hist(m, values, self._asimov_bins) for m in model]):
                weights, bin_edges = ad
                bin_centers = bin_edges[0: -1] + np.diff(bin_edges)/2
                asimov_data.append(array2dataset(type(data[i]), data[i].space, bin_centers, weights))

            self._asimov_dataset[poi] = asimov_data

        return self._asimov_dataset[poi]

    def asimov_loss(self, poi):
        """ Construct a loss function using the Asimov dataset for a given alternative hypothesis.

            Args:
                poi (`POI`): parameter of interest of the alternative hypothesis

            Returns:
                 Loss function

            Example with `zfit`:
                >>> poialt = POI(mean, 1.2)
                >>> loss = calc.asimov_loss([poialt])

        """
        if poi not in self._asimov_loss.keys():
            loss = self.lossbuilder(self.model, self.asimov_dataset(poi))
            self._asimov_loss[poi] = loss

        return self._asimov_loss[poi]

    def asimov_nll(self, pois, poialt) -> np.array:
        """ Compute negative log-likelihood values for given parameters of interest using the Asimov dataset
            generated with a given alternative hypothesis.

            Args:
                pois (`POIarray`): parameters of interest
                poialt (`POIarray`): parameter of interest of the alternative hypothesis

            Returns:
                 `numpy.array`: alternative nll values

            Example with `zfit`:
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> poinull = POIarray(mean, [1.1, 1.2, 1.0])
                >>> poialt = POI(mean, 1.2)
                >>> nll = calc.asimov_nll([poinull], [poialt])

        """
        self.check_pois(pois)
        self.check_pois(poialt)

        minimizer = self.minimizer
        ret = np.empty(pois.shape)
        for i, p in enumerate(pois):
            if p not in self._asimov_nll.keys():
                loss = self.asimov_loss(poialt)
                nll = pll(minimizer, loss, p)
                self._asimov_nll[p] = nll
            ret[i] = self._asimov_nll[p]
        return ret

    def pnull(self, qobs, qalt=None, onesided=True, onesideddiscovery=False, qtilde=False, nsigma=0) -> np.array:
        """ Compute the pvalue for the null hypothesis.

            Args:
                qobs (`np.array`): observed values of the test-statistic q
                qalt (`np.array`): alternative values of the test-statistic q using the asimov dataset
                onesided (bool, optional): if `True` (default) computes onesided pvalues
                onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery
                qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                    the $$q$$ test statistic
                nsigma (float, optional): significance shift

            Returns:
                 `np.array` : array of the pvalues for the null hypothesis
        """
        sqrtqobs = np.sqrt(qobs)

        # 1 - norm.cdf(x) == norm.cdf(-x)
        if onesided or onesideddiscovery:
            pnull = 1. - norm.cdf(sqrtqobs - nsigma)
        else:
            pnull = (1. - norm.cdf(sqrtqobs - nsigma))*2.

        if qalt is not None and qtilde:
            cond = (qobs > qalt) & (qalt > 0)
            sqrtqalt = np.sqrt(qalt)
            pnull_2 = 1. - norm.cdf((qobs + qalt) / (2. * sqrtqalt) - nsigma)

            if not (onesided or onesideddiscovery):
                pnull_2 += 1. - norm.cdf(sqrtqobs - nsigma)

            pnull = np.where(cond, pnull_2, pnull)

        return pnull

    def qalt(self, poinull, poialt, onesided, onesideddiscovery) -> np.array:
        """ Compute alternative hypothesis values of the $$\\Delta$$ log-likelihood test statistic using the asimov
            dataset.

            Args:
                poinull (`POIarray`): parameters of interest for the null hypothesis
                poialt (`POIarray`): parameters of interest for the alternative hypothesis
                onesided (bool, optional): if `True` (default) computes onesided pvalues
                onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery
                    test

            Returns:
                `numpy.array`: observed values of q

            Example with `zfit`:
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> poinull = POI(mean, [1.1, 1.2, 1.0])
                >>> poialt = POI(mean, [1.2])
                >>> q = calc.qalt([poinull], [poialt])
        """
        nll_poinull_asy = self.asimov_nll(poinull, poialt)
        nll_poialt_asy = self.asimov_nll(poialt, poialt)
        return self.q(nll1=nll_poinull_asy, nll2=nll_poialt_asy, poi1=poinull, poi2=poialt,
                      onesided=onesided, onesideddiscovery=onesideddiscovery)

    def palt(self, qobs, qalt, onesided=True, onesideddiscovery=False, qtilde=False) -> np.array:
        """ Compute the pvalue for the alternative hypothesis.

            Args:
                qobs (`np.array`): observed values of the test-statistic q
                qalt (`np.array`): alternative values of the test-statistic q using the Asimov dataset
                onesided (bool, optional): if `True` (default) computes onesided pvalues
                onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery
                qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                    the $$q$$ test statistic

            Returns:
                 `np.array` : array of the pvalues for the alternative hypothesis
        """
        sqrtqobs = np.sqrt(qobs)
        sqrtqalt = np.sqrt(qalt)

        # 1 - norm.cdf(x) == norm.cdf(-x)
        if onesided or onesideddiscovery:
            palt = 1. - norm.cdf(sqrtqobs - sqrtqalt)
        else:
            palt = 1. - norm.cdf(sqrtqobs + sqrtqalt)
            palt += 1. - norm.cdf(sqrtqobs - sqrtqalt)

        if qtilde:
            cond = (qobs > qalt) & (qalt > 0)
            palt_2 = 1. - norm.cdf((qobs - qalt) / (2. * sqrtqalt))

            if not (onesided or onesideddiscovery):
                palt_2 += 1. - norm.cdf(sqrtqobs + sqrtqalt)

            palt = np.where(cond, palt_2, palt)

        return palt

    def _pvalue_(self, poinull, poialt, qtilde, onesided, onesideddiscovery):

        qobs = self.qobs(poinull, onesided=onesided, qtilde=qtilde,
                         onesideddiscovery=onesideddiscovery)

        needpalt = poialt is not None

        if needpalt:
            qalt = self.qalt(poinull, poialt, onesided, onesideddiscovery)
            palt = self.palt(qobs=qobs, qalt=qalt, onesided=onesided, qtilde=qtilde,
                             onesideddiscovery=onesideddiscovery)
        else:
            qalt = None
            palt = None

        pnull = self.pnull(qobs=qobs, qalt=qalt, onesided=onesided, qtilde=qtilde,
                           onesideddiscovery=onesideddiscovery)

        return pnull, palt

    def _expected_pvalue_(self, poinull, poialt, nsigma, CLs, onesided, onesideddiscovery, qtilde):

        qalt = self.qalt(poinull, poialt, onesided=onesided, onesideddiscovery=onesideddiscovery)
        qalt = np.where(qalt < 0, 0, qalt)

        expected_pvalues = []
        for ns in nsigma:
            p_clsb = self.pnull(qobs=qalt, qalt=None, onesided=onesided, qtilde=qtilde,
                                onesideddiscovery=onesideddiscovery, nsigma=ns)
            if CLs:
                p_clb = norm.cdf(ns)
                p_cls = p_clsb / p_clb
                expected_pvalues.append(np.where(p_cls < 0, 0, p_cls))
            else:
                expected_pvalues.append(np.where(p_clsb < 0, 0, p_clsb))

        return expected_pvalues

    def _expected_poi_(self, poinull, poialt, nsigma, alpha, CLs, onesided, onesideddiscovery):

        qalt = self.qalt(poinull, poialt, onesided=onesided, onesideddiscovery=onesideddiscovery)
        qalt = np.where(qalt < 0, 0, qalt)

        sigma = np.sqrt((poinull[0].value - poialt[0].value)**2 / qalt)

        expected_values = []
        for ns in nsigma:
            if CLs:
                exp = sigma * (norm.ppf(1 - alpha*norm.cdf(ns)) + ns)
            else:
                exp = sigma * (norm.ppf(1 - alpha) + ns)

            expected_values.append(poialt[0].value + exp)

        return expected_values



from scipy.stats import norm

from .basetest import BaseTest


class Discovery(BaseTest):
    def __init__(self, calculator, poinull):
        """Class for discovery test.

            Args:
                calculator (`sktats.hypotests.BaseCalculator`): calculator to use for computing the pvalues
                poinull (`POI`): parameter of interest for the null hypothesis

            Example with `zfit`:
                >>> import zfit
                >>> from zfit.loss import ExtendedUnbinnedNLL
                >>> from zfit.minimize import MinuitMinimizer

                >>> bounds = (0.1, 3.0)
                >>> zfit.Space('x', limits=bounds)

                >>> bkg = np.random.exponential(0.5, 300)
                >>> peak = np.random.normal(1.2, 0.1, 25)
                >>> data = np.concatenate((bkg, peak))
                >>> data = data[(data > bounds[0]) & (data < bounds[1])]
                >>> N = data.size
                >>> data = zfit.data.Data.from_numpy(obs=obs, array=data)

                >>> lambda_ = zfit.Parameter("lambda", -2.0, -4.0, -1.0)
                >>> Nsig = zfit.Parameter("Ns", 20., -20., N)
                >>> Nbkg = zfit.Parameter("Nbkg", N, 0., N*1.1)
                >>> signal = Nsig * zfit.pdf.Gauss(obs=obs, mu=1.2, sigma=0.1)
                >>> background = Nbkg * zfit.pdf.Exponential(obs=obs, lambda_=lambda_)
                >>> loss = ExtendedUnbinnedNLL(model=[signal + background], data=[data])

                >>> from hepstats.hypotests.calculators import AsymptoticCalculator
                >>> from hepstats.hypotests import Discovery
                >>> from hepstats.hypotests.parameters import POI

                >>> calculator = AsymptoticCalculator(loss, MinuitMinimizer())
                >>> poinull = POI(Nsig, 0)
                >>> discovery_test = Discovery(calculator, [poinull])
                >>> discovery_test.result()
                p_value for the Null hypothesis = 0.0007571045424956679
                Significance (in units of sigma) = 3.1719464825102244
        """
        super(Discovery, self).__init__(calculator, poinull)

    def result(self, printlevel=1):
        """
        Returns the result of the discovery hypothesis test.

            Args:
                printlevel (int, default=1): if > 0 print the result

            Returns:
                Tuple(float, float): pnull, significance
        """
        pnull, _ = self.calculator.pvalue(self.poinull, onesideddiscovery=True)
        pnull = pnull[0]

        significance = norm.ppf(1. - pnull)

        if printlevel > 0:
            print("\np_value for the Null hypothesis = {0}".format(pnull))
            print("Significance (in units of sigma) = {0}".format(significance))

        return pnull, significance



#!/usr/bin/python
import numpy as np
from collections.abc import Iterable

from .fitutils.api_check import is_valid_parameter


class POIarray(object):

    def __init__(self, parameter, values):
        """
        Class for parameters of interest with multiple values:

            Args:
                parameter: the parameter of interest
                values (list(float)`,`numpy.array`): values of the parameter of interest

            Example with `zfit`:
                >>> Nsig = zfit.Parameter("Nsig")
                >>> poi = POIarray(Nsig, value=np.linspace(0,10,10))
        """

        if not is_valid_parameter(parameter):
            raise ValueError(f"{parameter} is not a valid parameter!")

        if not isinstance(values, Iterable):
            raise TypeError("A list/array of values of the POI is required.")

        self.parameter = parameter
        self.name = parameter.name
        self._values = np.array(values, dtype=np.float64)
        self._ndim = 1
        self._shape = (len(values),)

    @property
    def values(self):
        """
        Returns the values of the `POIarray`.
        """
        return self._values

    def __repr__(self):
        return "POIarray('{0}', values={1})".format(self.name, self.values)

    def __getitem__(self, i):
        """
        Get the i th element the array of values of the `POIarray`.
        """
        return POI(self.parameter, self.values[i])

    def __iter__(self):
        for v in self.values:
            yield POI(self.parameter, v)

    def __len__(self):
        return len(self.values)

    def __eq__(self, other):
        if not isinstance(other, POIarray):
            return NotImplemented

        if len(self) != len(other):
            return False

        values_equal = self.values == other.values
        name_equal = self.name == other.name
        return values_equal.all() and name_equal

    def __hash__(self):
        return hash((self.name, self.values.tostring()))

    @property
    def ndim(self):
        return self._ndim

    @property
    def shape(self):
        return self._shape

    def append(self, values):
        if not isinstance(values, Iterable):
            values = [values]
        values = np.concatenate([self.values, values])
        return POIarray(parameter=self.parameter, values=values)


class POI(POIarray):

    def __init__(self, parameter, value):
        """
        Class for single value parameters of interest:

            Args:
                parameter: the parameter of interest
                values (`float`,`int?): value of the parameter of interest

            Example with `zfit`:
                >>> Nsig = zfit.Parameter("Nsig")
                >>> poi = POI(Nsig, value=0)
        """
        if isinstance(value, Iterable):
            raise TypeError("A single value for the POI is required.")

        super(POI, self).__init__(parameter=parameter, values=[value])
        self._value = value

    @property
    def value(self):
        """
        Returns the value of the `POI`.
        """
        return self._value

    def __eq__(self, other):
        if not isinstance(other, POI):
            return NotImplemented

        value_equal = self.value == other.value
        name_equal = self.name == other.name
        return value_equal and name_equal

    def __repr__(self):
        return "POI('{0}', value={1})".format(self.name, self.value)

    def __hash__(self):
        return hash((self.name, self.value))


def asarray(POI):
    return POIarray(POI.parameter, POI.values)



from .fitutils.api_check import is_valid_loss, is_valid_fitresult, is_valid_minimizer
from .fitutils.api_check import is_valid_data, is_valid_pdf
from .fitutils.utils import get_nevents


class HypotestsObject(object):

    def __init__(self, input, minimizer):
        """Base object in `hepstats.hypotests` to manipulate a loss function and a minimizer.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function
        """

        if is_valid_fitresult(input):
            self._loss = input.loss
            self._bestfit = input
        elif is_valid_loss(input):
            self._loss = input
            self._bestfit = None
        else:
            raise ValueError("{} is not a valid loss funtion or fit result!".format(input))

        if not is_valid_minimizer(minimizer):
            raise ValueError("{} is not a valid minimizer !".format(minimizer))

        self._minimizer = minimizer
        self.minimizer.verbosity = 0

        self._parameters = {}
        for m in self.model:
            for d in m.get_dependents():
                self._parameters[d.name] = d

    @property
    def loss(self):
        """
        Returns the loss / likelihood function.
        """
        return self._loss

    @property
    def minimizer(self):
        """
        Returns the minimizer.
        """
        return self._minimizer

    @property
    def bestfit(self):
        """
        Returns the best fit values of the model parameters.
        """
        if getattr(self, "_bestfit", None):
            return self._bestfit
        else:
            print("Get fit best values!")
            self.minimizer.verbosity = 5
            mininum = self.minimizer.minimize(loss=self.loss)
            self.minimizer.verbosity = 0
            self._bestfit = mininum
            return self._bestfit

    @bestfit.setter
    def bestfit(self, value):
        """
        Set the best fit values  of the model parameters.

            Args:
                value: fit result
        """
        if not is_valid_fitresult(value):
            raise ValueError()
        self._bestfit = value

    @property
    def model(self):
        """
        Returns the model.
        """
        return self.loss.model

    @property
    def data(self):
        """
        Returns the data.
        """
        return self.loss.data

    @property
    def constraints(self):
        """
        Returns the constraints on the loss / likehood function.
        """
        return self.loss.constraints

    def get_parameter(self, name):
        """
        Returns the parameter in loss function with given input name.

        Args:
            name (str): name of the parameter to return
        """
        return self._parameters[name]

    def set_dependents_to_bestfit(self):
        """
        Set the values of the parameters in the models to the best fit values
        """
        for m in self.model:
            for d in m.get_dependents():
                d.set_value(self.bestfit.params[d]["value"])

    def lossbuilder(self, model, data, weights=None):
        """ Method to build a new loss function.

            Args:
                model (List): The model or models to evaluate the data on
                data (List): Data to use
                weights (optional, List): the data weights

            Example with `zfit`:
                >>> data = zfit.data.Data.from_numpy(obs=obs, array=np.random.normal(1.2, 0.1, 10000))
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> sigma = zfit.Parameter("sigma", 0.1)
                >>> model = zfit.pdf.Gauss(obs=obs, mu=mean, sigma=sigma)
                >>> loss = calc.lossbuilder(model, data)

            Returns:
                Loss function

        """

        assert all(is_valid_pdf(m) for m in model)
        assert all(is_valid_data(d) for d in data)

        msg = "{0} must have the same number of components as {1}"
        if len(data) != len(self.data):
            raise ValueError(msg.format("data", "`self.data"))
        if len(model) != len(self.model):
            raise ValueError(msg.format("model", "`self.model"))
        if weights is not None and len(weights) != len(self.data):
            raise ValueError(msg.format("weights", "`self.data`"))

        if weights is not None:
            for d, w in zip(data, weights):
                d.set_weights(w)

        loss = type(self.loss)(model=model, data=data)
        loss.add_constraints(self.constraints)

        return loss


class ToysObject(HypotestsObject):

    def __init__(self, input, minimizer, sampler, sample):
        """Base object in `hepstats.hypotests` to manipulate a loss function, a minimizer and sample a
        model (within the loss function) to do toy experiments.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function
                sampler : function used to create sampler with models, number of events and
                    floating parameters in the sample.
                sample : function used to get samples from the sampler.
        """
        super(ToysObject, self).__init__(input, minimizer)
        self._toys = {}
        self._sampler = sampler
        self._sample = sample
        self._toys_loss = {}

    def sampler(self, floating_params=None):
        """
        Create sampler with models.

        Args:
            floating_params (list): floating parameters in the sampler

        Example with `zfit`:
            >>> sampler = calc.sampler(floating_params=[zfit.Parameter("mean")])
        """
        self.set_dependents_to_bestfit()
        nevents = []
        for m, d in zip(self.loss.model, self.loss.data):
            if m.is_extended:
                nevents.append("extended")
            else:
                nevents.append(get_nevents(d))

        return self._sampler(self.loss.model,  nevents, floating_params)

    def sample(self, sampler, ntoys, poi=None):
        """
        Returns the samples generated from the sampler for a given value of a parameter of interest

        Args:
            sampler (list): generator of samples
            ntoys (int): number of samples to generate
            poi (POI, optional):  in the sampler

        Example with `zfit`:
            >>> mean = zfit.Parameter("mean")
            >>> sampler = calc.sampler(floating_params=[mean])
            >>> sample = calc.sample(sampler, 1000, POI(mean, 1.2))
        """
        return self._sample(sampler, ntoys, parameter=poi.parameter, value=poi.value)

    def toys_loss(self, parameter_name):
        """
        Construct a loss function constructed with a sampler for a given floating parameter

        Args:
            parameter_name: name floating parameter in the sampler
        Returns:
             Loss function

        Example with `zfit`:
            >>> loss = calc.toys_loss(zfit.Parameter("mean"))
        """
        if parameter_name not in self._toys_loss:
            parameter = self.get_parameter(parameter_name)
            sampler = self.sampler(floating_params=[parameter])
            self._toys_loss[parameter.name] = self.lossbuilder(self.model, sampler)
        return self._toys_loss[parameter_name]



"""
Module for testing a fitting library validity with hepstats.

A fitting library should provide six basic objects:

    * model / probability density function
    * parameters of the models
    * data
    * loss / likelihood function
    * minimizer
    * fitresult (optional)

A function for each object is defined in this module, all should return `True` to work
with hepstats.

The `zfit` API is currently the standard fitting API in hepstats.

"""


def is_valid_parameter(object):
    has_value = hasattr(object, "value")
    has_set_value = hasattr(object, "set_value")
    has_floating = hasattr(object, "floating")

    return has_value and has_set_value and has_floating


def is_valid_data(object):
    is_sampled_data = hasattr(object, "resample")

    try:
        has_nevents = hasattr(object, "nevents")
    except RuntimeError:
        if is_sampled_data:
            object.resample()
            has_nevents = hasattr(object, "nevents")
        else:
            has_nevents = False

    has_weights = hasattr(object, "weights")
    has_set_weights = hasattr(object, "set_weights")
    has_space = hasattr(object, "space")
    return has_nevents and has_weights and has_set_weights and has_space


def is_valid_pdf(object):
    has_get_dependents = hasattr(object, "get_dependents")
    if not has_get_dependents:
        return False
    else:
        params = object.get_dependents()

    all_valid_params = all(is_valid_parameter(p) for p in params)
    has_pdf = hasattr(object, "pdf")
    has_integrate = hasattr(object, "integrate")
    has_sample = hasattr(object, "sample")
    has_space = hasattr(object, "space")
    has_get_yield = hasattr(object, "get_yield")

    return all_valid_params and has_pdf and has_integrate and has_sample and has_space and has_get_yield


def is_valid_loss(object):
    if not hasattr(object, "model"):
        return False
    else:
        model = object.model

    if not hasattr(object, "data"):
        return False
    else:
        data = object.data

    has_constraints = hasattr(object, "constraints")
    has_fit_range = hasattr(object, "fit_range")
    all_valid_pdfs = all(is_valid_pdf(m) for m in model)
    all_valid_datasets = all(is_valid_data(d) for d in data)

    return all_valid_pdfs and all_valid_datasets and has_constraints and has_fit_range


def is_valid_fitresult(object):
    has_loss = hasattr(object, "loss")

    if not has_loss:
        return False
    else:
        loss = object.loss
        has_params = hasattr(object, "params")
        return is_valid_loss(loss) and has_params


def is_valid_minimizer(object):
    has_minimize = hasattr(object, "minimize")
    return has_minimize



from scipy import interpolate

from .basetest import BaseTest
from ..exceptions import POIRangeError


class ConfidenceInterval(BaseTest):
    def __init__(self, calculator, poinull, qtilde=False):
        """Class for confidence interval calculation.

            Args:
                calculator (`sktats.hypotests.BaseCalculator`): calculator to use for computing the pvalues
                poinull (`POIarray`): parameters of interest for the null hypothesis
                qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                    the $$q$$ test statistic

            Example with `zfit`:
                >>> import numpy as np
                >>> import zfit
                >>> from zfit.loss import ExtendedUnbinnedNLL
                >>> from zfit.minimize import MinuitMinimizer

                >>> bounds = (0.1, 3.0)
                >>> zfit.Space('x', limits=bounds)

                >>> bkg = np.random.exponential(0.5, 300)
                >>> peak = np.random.normal(1.2, 0.1, 80)
                >>> data = np.concatenate((bkg, peak))
                >>> data = data[(data > bounds[0]) & (data < bounds[1])]
                >>> N = data.size
                >>> data = zfit.data.Data.from_numpy(obs=obs, array=data)

                >>> mean = zfit.Parameter("mean", 1.2, 0.5, 2.0)
                >>> sigma = zfit.Parameter("sigma", 0.1, 0.02, 0.2)
                >>> lambda_ = zfit.Parameter("lambda", -2.0, -4.0, -1.0)
                >>> Nsig = zfit.Parameter("Ns", 20., -20., N)
                >>> Nbkg = zfit.Parameter("Nbkg", N, 0., N*1.1)
                >>> signal = Nsig * zfit.pdf.Gauss(obs=obs, mu=mean, sigma=sigma)
                >>> background = Nbkg * zfit.pdf.Exponential(obs=obs, lambda_=lambda_)
                >>> loss = ExtendedUnbinnedNLL(model=[signal + background], data=[data])

                >>> from hepstats.hypotests.calculators import AsymptoticCalculator
                >>> from hepstats.hypotests import ConfidenceInterval
                >>> from hepstats.hypotests.parameters import POI, POIarray

                >>> calculator = AsymptoticCalculator(loss, MinuitMinimizer())
                >>> poinull = POIarray(mean, np.linspace(1.15, 1.26, 100))
                >>> ci = ConfidenceInterval(calculator, [poinull])
                >>> ci.interval()
                Confidence interval on mean:
                    1.1810371356602791 < mean < 1.2156701172321935 at 68.0% C.L.
        """

        super(ConfidenceInterval, self).__init__(calculator, poinull)

        self._qtilde = qtilde

    @property
    def qtilde(self):
        """
        Returns True if qtilde test statistic is used, else False.
        """
        return self._qtilde

    def pvalues(self):
        """
        Returns p-values scanned for the values of the parameters of interest
        in the null hypothesis.

        Returns:
            pvalues (`np.array`): CLsb, CLs, expected (+/- sigma bands) p-values
        """

        return self.calculator.pvalue(self.poinull, qtilde=self.qtilde, onesided=False)[0]

    def interval(self, alpha=0.32, printlevel=1):
        """
        Returns the confidence level on the parameter of interest.

        Args:
            alpha (float, default=0.32/1 sigma): significance level,
            printlevel (int, default=1): if > 0 print the result

        Returns:
            limits (Dict): central, upper and lower bounds on the parameter of interest

        """

        poinull = self.poinull
        observed = self.calculator.bestfit.params[poinull.parameter]["value"]

        if min(self.pvalues()) > alpha:
            msg = f"The minimum of the scanned p-values is {min(self.pvalues())} which is larger than the"
            msg += f" confidence level alpha = {alpha}. Try to increase the range of POI values."
            raise POIRangeError(msg)

        tck = interpolate.splrep(poinull.values, self.pvalues()-alpha, s=0)
        root = interpolate.sproot(tck)

        bands = {}
        bands["observed"] = observed

        if len(root) < 2:
            msg = f" bound on the POI `{poinull.name}` cannot not be interpolated."
            if root[0] < observed:
                msg = "Upper" + msg + " Try to increase the maximum POI value."
            else:
                msg = "Low" + msg + " Try to decrease the minimum POI value."
            raise POIRangeError(msg)

        else:
            bands["lower"] = root[0]
            bands["upper"] = root[1]

        if printlevel > 0:

            msg = f"\nConfidence interval on {poinull.name}:\n"
            msg += f"\t{root[0]} < {poinull.name} < {root[1]} at {(1-alpha)*100:.1f}% C.L."
            print(msg)

        return bands



import numpy as np
from scipy.stats import norm

from .basecalculator import ToysCalculator
from ..fitutils.sampling import base_sampler, base_sample
from ..parameters import POI, POIarray


class FrequentistCalculator(ToysCalculator):
    """
    Class for frequentist calculators.
    """

    def __init__(self, input, minimizer, ntoysnull=100, ntoysalt=100, sampler=base_sampler, sample=base_sample):
        """Frequentist calculator class.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function
                ntoysnull (int, default=100): minimum number of toys to generate for the null hypothesis
                ntoysalt (int, default=100): minimum number of toys to generate for the alternative hypothesis
                sampler : function used to create sampler with models, number of events and
                    floating parameters in the sample. Default is `hepstats.fitutils.sampling.base_sampler`.
                sample : function used to get samples from the sampler.
                    Default is `hepstats.fitutils.sampling.base_sample`.

            Example with `zfit`:
                >>> import zfit
                >>> from zfit.core.loss import UnbinnedNLL
                >>> from zfit.minimize import MinuitMinimizer

                >>> obs = zfit.Space('x', limits=(0.1, 2.0))
                >>> data = zfit.data.Data.from_numpy(obs=obs, array=np.random.normal(1.2, 0.1, 10000))
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> sigma = zfit.Parameter("sigma", 0.1)
                >>> model = zfit.pdf.Gauss(obs=obs, mu=mean, sigma=sigma)
                >>> loss = UnbinnedNLL(model=[model], data=[data], fit_range=[obs])

                >>> calc = FrequentistCalculator(input=loss, minimizer=MinuitMinimizer(), ntoysnull=1000, ntoysalt=1000)
        """

        super(FrequentistCalculator, self).__init__(input, minimizer, ntoysnull, ntoysalt, sampler, sample)

    def qnull(self, poinull, poialt, onesided, onesideddiscovery, qtilde=False):
        """ Compute null hypothesis values of the $$\\Delta$$ log-likelihood test statistic.

            Args:
                poinull (`POIarray`): parameters of interest for the null hypothesis
                poialt (`POIarray`): parameters of interest for the alternative hypothesis
                onesided (bool): if `True` computes onesided pvalues
                onesideddiscovery (bool): if `True` computes onesided pvalues for a discovery
                    test
                qtilde (bool): if `True` use the $$\tilde{q}$$ test statistics else use
                    the $$q$$ test statistic

            Returns:
                `numpy.array`: observed values of q

            Example with `zfit`:
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> poinull = POIarray(mean, [1.1, 1.2, 1.0])
                >>> poialt = POI(mean, 1.2)
                >>> q = calc.qnull(poinull, poialt)
        """
        toysresults = self.get_toys_null(poinull, poialt, qtilde)
        ret = {}

        for p in poinull:
            toysresult = toysresults[p]
            nll1 = toysresult.nlls[p]
            nll2 = toysresult.nll_bestfit
            bestfit = toysresult.bestfit

            if qtilde:
                nllat0 = toysresult.nlls[POI(poinull.parameter, 0.0)]
                nll2 = np.where(bestfit < 0, nllat0, nll2)
                bestfit = np.where(bestfit < 0, 0, bestfit)

            poi1 = POIarray(poinull.parameter, np.full(nll1.size, p.value))
            poi2 = POIarray(poinull.parameter, bestfit)

            ret[p] = self.q(nll1=nll1, nll2=nll2, poi1=poi1, poi2=poi2, onesided=onesided,
                            onesideddiscovery=onesideddiscovery)

        return ret

    def qalt(self, poinull, poialt, onesided, onesideddiscovery, qtilde=False):
        """ Compute alternative hypothesis values of the $$\\Delta$$ log-likelihood test statistic.

            Args:
                poinull (`POIarray`): parameters of interest for the null hypothesis
                poialt (`POIarray`): parameters of interest for the alternative hypothesis
                onesided (bool): if `True` computes onesided pvalues
                onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery
                    test
                qtilde (bool): if `True` use the $$\tilde{q}$$ test statistics else use
                    the $$q$$ test statistic

            Returns:
                `numpy.array`: observed values of q

            Example with `zfit`:
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> poinull = POIarray(mean, [1.1, 1.2, 1.0])
                >>> poialt = POI(mean, 1.2)
                >>> q = calc.qalt(poinull, poialt)
        """
        toysresults = self.get_toys_alt(poialt, poinull, qtilde)
        ret = {}

        for p in poinull:
            toysresult = toysresults[poialt]

            nll1 = toysresult.nlls[p]
            nll2 = toysresult.nll_bestfit
            bestfit = toysresult.bestfit

            if qtilde:
                nllat0 = toysresult.nlls[POI(poinull.parameter, 0.0)]
                nll2 = np.where(bestfit < 0, nllat0, nll2)
                bestfit = np.where(bestfit < 0, 0, bestfit)

            poi1 = POIarray(poialt.parameter, np.full(nll1.size, p.value))
            poi2 = POIarray(poialt.parameter, bestfit)

            ret[p] = self.q(nll1=nll1, nll2=nll2, poi1=poi1, poi2=poi2, onesided=onesided,
                            onesideddiscovery=onesideddiscovery)

        return ret

    def _pvalue_(self, poinull, poialt, qtilde, onesided, onesideddiscovery):

        qobs = self.qobs(poinull, onesided=onesided, qtilde=qtilde,
                         onesideddiscovery=onesideddiscovery)

        def compute_pvalue(qdist, qobs):
            qdist = qdist[~(np.isnan(qdist) | np.isinf(qdist))]
            p = len(qdist[qdist >= qobs])/len(qdist)
            return p

        qnulldist = self.qnull(poinull, poialt, onesided, onesideddiscovery, qtilde)
        pnull = np.empty(len(poinull))
        for i, p in enumerate(poinull):
            pnull[i] = compute_pvalue(qnulldist[p], qobs[i])

        if poialt is not None:
            qaltdist = self.qalt(poinull, poialt, onesided, onesideddiscovery, qtilde)
            palt = np.empty(len(poinull))
            for i, p in enumerate(poinull):
                palt[i] = compute_pvalue(qaltdist[p], qobs[i])
        else:
            palt = None

        return pnull, palt

    def _expected_pvalue_(self, poinull, poialt, nsigma, CLs, onesided, onesideddiscovery, qtilde):

        ps = {ns: {"p_clsb": np.empty(len(poinull)),
                   "p_clb": np.empty(len(poinull))} for ns in nsigma}

        qnulldist = self.qnull(poinull, poialt, onesided, onesideddiscovery, qtilde)
        qaltdist = self.qalt(poinull, poialt, onesided, onesideddiscovery, qtilde)

        filter_nan = lambda q: q[~(np.isnan(q) | np.isinf(q))]

        for i, p in enumerate(poinull):

            qaltdist_p = filter_nan(qaltdist[p])
            lqaltdist = len(qaltdist_p)

            qnulldist_p = filter_nan(qnulldist[p])
            lqnulldist = len(qnulldist_p)

            p_clsb_i = np.empty(lqaltdist)
            p_clb_i = np.empty(lqaltdist)

            for j, q in np.ndenumerate(qaltdist_p):
                p_clsb_i[j] = (len(qnulldist_p[qnulldist_p >= q])/lqnulldist)
                p_clb_i[j] = (len(qaltdist_p[qaltdist_p >= q])/lqaltdist)

            for ns in nsigma:
                frac = norm.cdf(ns)*100
                ps[ns]["p_clsb"][i] = np.percentile(p_clsb_i, frac)
                ps[ns]["p_clb"][i] = np.percentile(p_clb_i, frac)

        expected_pvalues = []
        for ns in nsigma:
            if CLs:
                p_cls = ps[ns]["p_clsb"] / ps[ns]["p_clb"]
                expected_pvalues.append(np.where(p_cls < 0, 0, p_cls))
            else:
                p_clsb = ps[ns]["p_clsb"]
                expected_pvalues.append(np.where(p_clsb < 0, 0, p_clsb))

        return expected_pvalues



#!/usr/bin/python
from typing import Dict, Union, Tuple, List
import numpy as np

from ..hypotests_object import HypotestsObject
from ..parameters import POI, POIarray, asarray
from ..fitutils.utils import pll
from ..toyutils import ToysManager
from ..fitutils.sampling import base_sampler, base_sample

"""
Module defining the base class for the calculators for statistical tests based on the likelyhood ratio.

Any calculator can be a subclass of `BaseCalculator`. Currently implemented:

    * `AsymptoticCalculator`: calculator using the asymptotic formulae of the likehood ratio.

Acronyms used in the code:
    * nll = negative log-likehood, the likehood being the `loss` attribute of a calculator;
    * obs = observed, i.e. measured on provided data.

"""


class BaseCalculator(HypotestsObject):

    def __init__(self, input, minimizer):
        """Base class for calculator.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function

            Example with `zfit`:
                >>> import zfit
                >>> from zfit.core.loss import UnbinnedNLL
                >>> from zfit.minimize import MinuitMinimizer

                >>> obs = zfit.Space('x', limits=(0.1, 2.0))
                >>> data = zfit.data.Data.from_numpy(obs=obs, array=np.random.normal(1.2, 0.1, 10000))
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> sigma = zfit.Parameter("sigma", 0.1)
                >>> model = zfit.pdf.Gauss(obs=obs, mu=mean, sigma=sigma)
                >>> loss = UnbinnedNLL(model=[model], data=[data], fit_range=[obs])

                >>> calc = BaseCalculator(input=loss, minimizer=MinuitMinimizer())
        """

        super(BaseCalculator, self).__init__(input, minimizer)

        self._obs_nll = {}

        self._parameters = {}
        for m in self.model:
            for d in m.get_dependents():
                self._parameters[d.name] = d

    def obs_nll(self, pois) -> np.array:
        """ Compute observed negative log-likelihood values for given parameters of interest.

            Args:
                pois (List[`hypotests.POI`]): parameters of interest

            Returns:
                 numpy.array`: observed nll values

            Example with `zfit`:
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> poi = POI(mean, [1.1, 1.2, 1.0])
                >>> nll = calc.obs_nll([poi])

        """
        ret = np.empty(pois.shape)
        for i, p in enumerate(pois):
            if p not in self._obs_nll.keys():
                nll = pll(minimizer=self.minimizer, loss=self.loss, pois=p)
                self._obs_nll[p] = nll
            ret[i] = self._obs_nll[p]
        return ret

    def qobs(self, poinull: List[POI], onesided=True, onesideddiscovery=False, qtilde=False):
        """ Compute observed values of the $$\\Delta$$ log-likelihood test statistic.

            Args:
                poinull (List[`hypotests.POI`]): parameters of interest for the null hypothesis
                qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                    the $$q$$ test statistic
                onesided (bool, optional): if `True` (default) computes onesided pvalues
                onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery
                    test

            Returns:
                `numpy.array`: observed values of q

            Example with `zfit`:
                >>> mean = zfit.Parameter("mu", 1.2)
                >>> poi = POI(mean, [1.1, 1.2, 1.0])
                >>> q = calc.qobs([poi])
        """

        self.check_pois(poinull)

        param = poinull.parameter
        bestfit = self.bestfit.params[param]["value"]
        if qtilde and poinull.ndim == 1:
            bestfitpoi = POI(param, 0)
        else:
            bestfitpoi = POI(param, bestfit)
            if len(poinull) == 1:
                self._obs_nll[bestfitpoi] = self.bestfit.fmin

        nll_poinull_obs = self.obs_nll(poinull)
        nll_bestfitpoi_obs = self.obs_nll(bestfitpoi)
        qobs = self.q(nll1=nll_poinull_obs, nll2=nll_bestfitpoi_obs, poi1=poinull, poi2=bestfitpoi,
                      onesided=onesided, onesideddiscovery=onesideddiscovery)

        return qobs

    def pvalue(self, poinull: List[POI], poialt: Union[List[POI], None] = None, qtilde=False, onesided=True,
               onesideddiscovery=False) -> Tuple[np.array, np.array]:
        """Computes pvalues for the null and alternative hypothesis.

        Args:
            poinull (List[`hypotests.POI`]): parameters of interest for the null hypothesis
            poialt (List[`hypotests.POI`], optional): parameters of interest for the alternative hypothesis
            qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                the $$q$$ test statistic
            onesided (bool, optional): if `True` (default) computes onesided pvalues
            onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery
                test

        Returns:
            Tuple(`numpy.array`, `numpy.array`): pnull, palt

        Example with `zfit`:
            >>> mean = zfit.Parameter("mu", 1.2)
            >>> poinull = POI(mean, [1.1, 1.2, 1.0])
            >>> poialt = POI(mean, 1.2)
            >>> pvalues = calc.pavalue([poinull], [poialt])
        """
        self.check_pois(poinull)
        if poialt:
            self.check_pois(poialt)
            self.check_pois_compatibility(poinull, poialt)

        return self._pvalue_(poinull=poinull, poialt=poialt, qtilde=qtilde, onesided=onesided,
                             onesideddiscovery=onesideddiscovery)

    def _pvalue_(self, poinull, poialt, qtilde, onesided, onesideddiscovery):
        """
        To be overwritten in `BaseCalculator` subclasses.
        """
        raise NotImplementedError

    def expected_pvalue(self, poinull: List[POI], poialt: List[POI], nsigma, CLs=False, qtilde=False,
                        onesided=True, onesideddiscovery=False) -> Dict[int, np.array]:
        """Computes the expected pvalues and error bands for different values of $$\\sigma$$ (0=expected/median)

        Args:
            poinull (List[`hypotests.POI`]): parameters of interest for the null hypothesis
            poialt (List[`hypotests.POI`], optional): parameters of interest for the alternative hypothesis
            nsigma (`numpy.array`): array of values of $$\\sigma$$ to compute the expected pvalue
            CLs (bool, optional): if `True` computes pvalues as $$p_{cls}=p_{null}/p_{alt}=p_{clsb}/p_{clb}$$
                else as $$p_{clsb} = p_{null}$
            qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                the $$q$$ test statistic
            onesided (bool, optional): if `True` (default) computes onesided pvalues
            onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery

        Returns:
            `numpy.array`: array of expected pvalues for each $$\\sigma$$ value

        Example with `zfit`:
            >>> mean = zfit.Parameter("mu", 1.2)
            >>> poinull = POI(mean, [1.1, 1.2, 1.0])
            >>> poialt = POI(mean, 1.2)
            >>> nll = calc.expected_pvalue([poinull], [poialt])
        """
        self.check_pois(poinull)
        if poialt:
            self.check_pois(poialt)
            self.check_pois_compatibility(poinull, poialt)

        return self._expected_pvalue_(poinull=poinull, poialt=poialt, nsigma=nsigma, CLs=CLs, qtilde=qtilde,
                                      onesided=onesided, onesideddiscovery=onesideddiscovery)

    def _expected_pvalue_(self, poinull, poialt, nsigma, CLs, qtilde, onesided, onesideddiscovery):
        """
        To be overwritten in `BaseCalculator` subclasses.
        """
        raise NotImplementedError

    def expected_poi(self, poinull: List[POI], poialt: List[POI], nsigma, alpha=0.05, CLs=False,
                     onesided=True, onesideddiscovery=False):
        """Computes the expected parameter of interest values such that the expected p_values == $$\alpha$$
        for different values of $$\\sigma$$ (0=expected/median)

        Args:
            poinull (List[`hypotests.POI`]): parameters of interest for the null hypothesis
            poialt (List[`hypotests.POI`], optional): parameters of interest for the alternative hypothesis
            nsigma (`numpy.array`): array of values of $$\\sigma$$ to compute the expected pvalue
            alpha (float, default=0.05): significance level
            CLs (bool, optional): if `True` uses pvalues as $$p_{cls}=p_{null}/p_{alt}=p_{clsb}/p_{clb}$$
                else as $$p_{clsb} = p_{null}$
            onesided (bool, optional): if `True` (default) computes onesided pvalues
            onesideddiscovery (bool, optional): if `True` (default) computes onesided pvalues for a discovery

        Returns:
            `numpy.array`: array of expected POI values for each $$\\sigma$$ value

        Example with `zfit`:
            >>> mean = zfit.Parameter("mu", 1.2)
            >>> poinull = POI(mean, [1.1, 1.2, 1.0])
            >>> poialt = POI(mean, 1.2)
            >>> nll = calc.expected_poi([poinull], [poialt])
        """
        self.check_pois(poinull)
        if poialt:
            self.check_pois(poialt)
            self.check_pois_compatibility(poinull, poialt)

        return self._expected_poi_(poinull=poinull, poialt=poialt, nsigma=nsigma, alpha=alpha, CLs=CLs,
                                   onesided=onesided, onesideddiscovery=onesideddiscovery)

    def _expected_poi_(self, poinull, poialt, nsigma, alpha, CLs, onesided, onesideddiscovery):
        """
        To be overwritten in `BaseCalculator` subclasses.
        """
        raise NotImplementedError

    @staticmethod
    def check_pois(pois):
        """
        Check if the parameters of interest are all `hepstats.parameters.POI/POIarray` instances.
        """

        msg = "POI/POIarray is required."
        if not isinstance(pois, POIarray):
            raise TypeError(msg)
        if pois.ndim > 1:
            msg = "Tests with more that one parameter of interest are not yet implemented."
            raise NotImplementedError(msg)

    @staticmethod
    def check_pois_compatibility(poi1, poi2):
        """
        Check compatibility between two lists of `hepstats.parameters.POIarray` instances.
        """

        if poi1.ndim != poi2.ndim:
            msg = f"POIs should have the same dimensions, poi1={poi1.ndim}, poi2={poi2.ndim}"
            raise ValueError(msg)

        if poi1.ndim == 1:

            if poi1.name != poi2.name:
                msg = "The variables used in the parameters of interest should have the same names,"
                msg += f" poi1={poi1.name}, poi2={poi2.name}"
                raise ValueError(msg)

    def q(self, nll1: np.array, nll2: np.array, poi1, poi2,
          onesided=True, onesideddiscovery=False) -> np.array:
        """ Compute value of the test statistic q defined as the difference between negative log-likelihood
            values $$q = nll1 - nll2$$

            Args:
                nll1 (`numpy.array`): array of nll values #1, evaluated with poi1
                nll2 (`numpy.array`): array of nll values #2, evaluated with poi2
                poi1 ((List[`hypotests.POI`])): list of POI's #1
                poi2 ((List[`hypotests.POI`])): list of POI's #2
                onesided (bool, optional, default=True)
                onesideddiscovery (bool, optional, default=True)

            Returns:
                `np.array`: array of q values
        """

        self.check_pois(poi1)
        self.check_pois(poi2)
        self.check_pois_compatibility(poi1, poi2)

        assert len(nll1) == len(poi1)
        assert len(nll2) == len(poi2)

        poi1 = poi1.values
        poi2 = poi2.values

        q = 2*(nll1 - nll2)
        # filter_non_nan = ~(np.isnan(q) | np.isinf(q))
        # q = q[filter_non_nan]
        #
        # if isinstance(poi2, np.ndarray):
        #     poi2 = poi2[filter_non_nan]
        zeros = np.zeros(q.shape)

        if onesideddiscovery:
            condition = (poi2 < poi1) | (q < 0)
            q = np.where(condition, zeros, q)
        elif onesided:
            condition = (poi2 > poi1) | (q < 0)
            q = np.where(condition, zeros, q)
        else:
            q = q

        return q


class BaseToysCalculator(BaseCalculator):

    def __init__(self, input, minimizer, sampler, sample):
        """Basis for toys calculator class.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function
                sampler : function used to create sampler with models, number of events and
                    floating parameters in the sample Default is `hepstats.fitutils.sampling.base_sampler`.
                sample : function used to get samples from the sampler.
                    Default is `hepstats.fitutils.sampling.base_sample`.
        """
        super(BaseToysCalculator, self).__init__(input, minimizer)


class ToysCalculator(BaseToysCalculator, ToysManager):
    """
    Class for calculators using toys.
    """

    def __init__(self, input, minimizer, ntoysnull=100, ntoysalt=100, sampler=base_sampler, sample=base_sample):
        """Toys calculator class.

            Args:
                input : loss or fit result
                minimizer : minimizer to use to find the minimum of the loss function
                ntoysnull (int, default=100): minimum number of toys to generate for the null hypothesis
                ntoysalt (int, default=100): minimum number of toys to generate for the alternative hypothesis
                sampler : function used to create sampler with models, number of events and
                    floating parameters in the sample Default is `hepstats.fitutils.sampling.base_sampler`.
                sample : function used to get samples from the sampler.
                    Default is `hepstats.fitutils.sampling.base_sample`.
        """
        super(ToysCalculator, self).__init__(input, minimizer, sampler, sample)

        self._ntoysnull = ntoysnull
        self._ntoysalt = ntoysalt

    @classmethod
    def from_yaml(cls, filename, loss, minimizer, ntoysnull=100, ntoysalt=100, sampler=base_sampler,
                  sample=base_sample):
        """
        ToysCalculator constructor with the toys loaded from a yaml file.

        Args:
            filename (str)
            input : loss or fit result
            minimizer : minimizer to use to find the minimum of the loss function
            ntoysnull (int, default=100): minimum number of toys to generate for the null hypothesis
            ntoysalt (int, default=100): minimum number of toys to generate for the alternative hypothesis
            sampler : function used to create sampler with models, number of events and
                floating parameters in the sample Default is `hepstats.fitutils.sampling.base_sampler`.
            sample : function used to get samples from the sampler.
                Default is `hepstats.fitutils.sampling.base_sample`.

        Returns
            ToysCalculator
        """

        calculator = cls(loss, minimizer, ntoysnull, ntoysalt, sampler, sample)
        toysresults = calculator.toysresults_from_yaml(filename)

        for t in toysresults:
            calculator.add_toyresult(t)

        return calculator

    @property
    def ntoysnull(self):
        """
        Returns the number of toys generated for the null hypothesis.
        """
        return self._ntoysnull

    @property
    def ntoysalt(self):
        """
        Returns the number of toys generated for the alternative hypothesis.
        """
        return self._ntoysalt

    def _get_toys(self, poigen, poieval=None, qtilde=False, hypothesis="null"):
        """
        Return the generated toys for a given POI.

        Args:
            poigen (POI): POI used to generate the toys
            poieval (POIarray): POI values to evaluate the loss function
            qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                the $$q$$ test statistic
            hypothesis: `null` or `alternative`
        """

        assert hypothesis in ["null", "alternative"]

        if hypothesis == "null":
            ntoys = self.ntoysnull
        else:
            ntoys = self.ntoysalt

        ret = {}

        for p in poigen:
            poieval_p = poieval

            if poieval_p is None:
                poieval_p = POIarray(poigen.parameter, [p.value])
            else:
                if p not in poieval_p:
                    poieval_p = poieval_p.append(p.value)

            if qtilde and 0. not in poieval_p.values:
                poieval_p = poieval_p.append(0.0)

            poieval_p = asarray(poieval_p)

            ngenerated = self.ntoys(p, poieval_p)
            if ngenerated < ntoys:
                ntogen = ntoys - ngenerated
            else:
                ntogen = 0

            if ntogen > 0:
                print(f"Generating {hypothesis} hypothesis toys for {p}.")

                self.generate_and_fit_toys(ntoys=ntogen, poigen=p, poieval=poieval_p)

            ret[p] = self.get_toyresult(p, poieval_p)

        return ret

    def get_toys_null(self, poigen, poieval, qtilde=False):
        """
        Return the generated toys for the null hypothesis.

        Args:
            poigen (POI): POI used to generate the toys
            ntoys (int): number of toys to generate
            poieval (POIarray): POI values to evaluate the loss function
            qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                the $$q$$ test statistic

        Example with `zfit`:
            >>> mean = zfit.Parameter("mu", 1.2)
            >>> poinull = POIarray(mean, [1.1, 1.2, 1.0])
            >>> poialt = POI(mean, 1.2)
            >>> for p in poinull:
            ...     calc.get_toys_alt(p, poieval=poialt)
        """
        return self._get_toys(poigen=poigen, poieval=poieval, qtilde=qtilde, hypothesis="null")

    def get_toys_alt(self, poigen, poieval, qtilde=False):
        """
        Return the generated toys for the alternative hypothesis.

        Args:
            poigen (POI): POI used to generate the toys
            ntoys (int): number of toys to generate
            poieval (POIarray): POI values to evaluate the loss function
            qtilde (bool, optional): if `True` use the $$\tilde{q}$$ test statistics else (default) use
                the $$q$$ test statistic

        Example with `zfit`:
            >>> mean = zfit.Parameter("mu", 1.2)
            >>> poinull = POIarray(mean, [1.1, 1.2, 1.0])
            >>> poialt = POI(mean, 1.2)
            >>> calc.get_toys_alt(poialt, poieval=poinull)
        """
        return self._get_toys(poigen=poigen, poieval=poieval, qtilde=qtilde, hypothesis="alternative")



from .api_check import is_valid_pdf

"""
Module providing basic sampling methods.
"""


def base_sampler(models, nevents, floating_params=None):
    """
    Creates samplers from models.

    Args:
        models (list(model)): models to sample
        nevents (list(int)): number of in each sampler
        floating_params (list(parameter), optionnal): floating parameter in the samplers
    """

    assert all(is_valid_pdf(m) for m in models)
    assert len(nevents) == len(models)

    if floating_params:
        floating_params_names = [f.name for f in floating_params]

    samplers = []
    fixed_params = []
    for m in models:
        def to_fix(p):
            if floating_params:
                return p.name in floating_params_names
            else:
                return False
        fixed = [p for p in m.get_dependents() if not to_fix(p)]
        fixed_params.append(fixed)

    for i, (m, p) in enumerate(zip(models, fixed_params)):
        sampler = m.create_sampler(n=nevents[i], fixed_params=p)
        samplers.append(sampler)

    return samplers


def base_sample(samplers, ntoys, parameter=None, value=None):
    """
    Sample from samplers. The parameters that are floating in the samplers can be set to a specific value
    using the `parameter` and `value` argument.

    Args:
        samplers (list): generators of samples
        ntoys (int): number of samples to generate
        parameter (optional): floating parameter in the sampler
        value (optional): value of the parameter
    """

    for i in range(ntoys):
        if not (parameter is None or value is None):
            with parameter.set_value(value):
                for s in samplers:
                    s.resample()
        else:
            for s in samplers:
                s.resample()

        yield i



from ..calculators.basecalculator import BaseCalculator

"""
Module defining the base class for hypothesis tests.
"""


class BaseTest(object):

    def __init__(self, calculator, poinull, poialt=None):
        """Base class for hypothesis tests.

            Args:
                calculator (`sktats.hypotests.BaseCalculator`): calculator to use for computing the pvalues
                poinull (List[`hypotests.POI`]): parameters of interest for the null hypothesis
                poialt (List[`hypotests.POI`], optional): parameters of interest for the alternative hypothesis
        """

        if not isinstance(calculator, BaseCalculator):
            msg = "Invalid type, {0}, for calculator. Calculator required."
            raise TypeError(msg)
        self._calculator = calculator

        self.calculator.check_pois(poinull)
        if poialt:
            self.calculator.check_pois(poialt)
            self.calculator.check_pois_compatibility(poinull, poialt)

        self._poinull = poinull
        self._poialt = poialt

    @property
    def poinull(self):
        """
        Returns the POI for the null hypothesis.
        """
        return self._poinull

    @property
    def poialt(self):
        """
        Returns the POI for the alternative hypothesis.
        """
        return self._poialt

    @property
    def calculator(self):
        """
        Returns the calculator.
        """
        return self._calculator



#!/usr/bin/env python
# Copyright (c) 2018-2019.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/hepstats for details.

from setuptools import setup
from setuptools import find_packages
import os

python_requires = "!=2.*, >=3.6"
install_requires = ["scipy", "numpy", "pandas", "asdf"]
setup_requires = ["pytest-runner"]
tests_requires = ["pytest", "zfit"]


def get_version():
    g = {}
    exec(open(os.path.join("hepstats", "version.py")).read(), g)
    return g["__version__"]


setup(
    name='hepstats',
    author='Matthieu Marinangeli',
    author_email='matthieu.marinangeli@cern.ch',
    maintainer='The Scikit-HEP admins',
    maintainer_email='scikit-hep-admins@googlegroups.com',
    version=get_version(),
    description='statistics tools and utilities',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/scikit-hep/hepstats',
    license='BSD 3-Clause License',
    packages=find_packages(),
    test_suite="tests",
    python_requires=python_requires,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_requires,
    keywords=[
            'HEP', 'statistics',
            ],
    classifiers=[
                'Topic :: Scientific/Engineering',
                'Intended Audience :: Science/Research',
                'Intended Audience :: Developers',
                'Operating System :: OS Independent',
                'License :: OSI Approved :: BSD License',
                'Programming Language :: Python',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8',
                'Development Status :: 4 - Beta'
                ],
    platforms="Any",
)



from contextlib import ExitStack
import numbers


def eval_pdf(model, x, params):
    """ Compute pdf of model at a given point x and for given parameters values """

    def pdf(model, x):
        if model.is_extended:
            ret = model.pdf(x) * model.get_yield()
        else:
            ret = model.pdf(x)

        if hasattr(ret, "numpy"):
            return ret.numpy()
        else:
            return ret

    with ExitStack() as stack:
        for param in model.get_dependents():
            value = params[param]["value"]
            stack.enter_context(param.set_value(value))
        return pdf(model, x)


def pll(minimizer, loss, pois) -> float:
    """ Compute minimum profile likelihood for given parameters values. """
    with ExitStack() as stack:
        for p in pois:
            param = p.parameter
            stack.enter_context(param.set_value(p.value))
            param.floating = False
        minimum = minimizer.minimize(loss=loss)
        for p in pois:
            p.parameter.floating = True
    return minimum.fmin


def array2dataset(dataset_cls, obs, array, weights=None):
    """
    dataset_cls: only used to get the class in which array/weights will be
    converted.
    """

    if hasattr(dataset_cls, "from_numpy"):
        return dataset_cls.from_numpy(obs=obs, array=array, weights=weights)
    else:
        return dataset_cls(obs=obs, array=array, weights=weights)


def get_nevents(dataset):
    """ Returns the number of events in the dataset """

    nevents = dataset.nevents

    if hasattr(nevents, "numpy"):
        return nevents.numpy()
    else:
        return nevents
